/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "queue.h"
#include "common.h"
#include "syncthread.h"

u64 get_block_size(struct s_blockinfo *blkinfo)
{
	return (blkinfo->blkcompressed==true)?(blkinfo->blkarsize):(blkinfo->blkrealsize);
}

struct timespec get_timeout()
{
	struct timespec t;
	clock_gettime(CLOCK_REALTIME, &t);
	t.tv_sec++;
	return t;
}

int queue_init(struct s_queue *q, s64 memmax)
{
	if (!q)
	{	errprintf("q is NULL\n");
		return -1;
	}
	q->head=NULL;
	q->curitemnum=1;
	q->memmax=memmax*1024*1024;
	q->memused=0;
	
	if (pthread_mutex_init(&q->mutex, NULL)!=0)
	{	errprintf("pthread_mutex_init failed\n");
		return -1;
	}
	if (pthread_cond_init(&q->cond,NULL)!=0)
	{	errprintf("pthread_cond_init failed\n");
		return -1;
	}
	
	return 0;
}

int queue_destroy(struct s_queue *q)
{
	struct s_queueitem *cur;
	struct s_queueitem *next;

	if (!q)
	{	errprintf("q is NULL\n");
		return -1;
	}
	
	pthread_mutex_destroy(&q->mutex);
	pthread_cond_destroy(&q->cond);
	
	cur=q->head;
	while (cur!=NULL)
	{
		next=cur->next;
		free(cur);
		cur=next;
	}
	q->head=NULL;
	return 0;
}

// add a block at the end of the queue
int queue_add_block(struct s_queue *q, struct s_blockinfo *blkinfo, int status)
{
	struct s_queueitem *item;
	struct s_queueitem *cur;
	
	if (!q || !blkinfo)
	{	errprintf("a parameter is NULL\n");
		return -1;
	}
	
	if (get_abort()==true)
		return -1;
	
	// create the new item in memory
	item=malloc(sizeof(struct s_queueitem));
	if (!item)
	{	errprintf("out of memory\n");
		return -1;
	}
	item->type=QITEM_TYPE_BLOCK;
	item->status=status;
	item->blkinfo=*blkinfo;
	item->next=NULL;
	q->memused+=get_block_size(blkinfo);
	//q->memused+=(blkinfo->blkcompressed==true)?(blkinfo->blkarsize):(blkinfo->blkrealsize);
	
	pthread_mutex_lock(&q->mutex);
	
	// if queue is full, let the other threads remove items first
	while (q->memused > q->memmax)
	{
		struct timespec t=get_timeout();
		pthread_cond_timedwait(&q->cond, &q->mutex, &t);
	}
	
	item->itemnum=q->curitemnum++;
	if (q->head==NULL) // if list empty: item is head
	{
		q->head=item;
	}
	else // list not empty: add items at the end
	{	for (cur=q->head; (cur!=NULL) && (cur->next!=NULL); cur=cur->next);
		cur->next=item;
	}
	
	pthread_mutex_unlock(&q->mutex);
	pthread_cond_broadcast(&q->cond);
	
	return 0;
}

// function called by the compression thread when a block has been compressed
int queue_replace_block(struct s_queue *q, s64 itemnum, struct s_blockinfo *blkinfo, int newstatus)
{
	struct s_queueitem *cur;
	u64 oldsize, newsize;
	
	if (!q || !blkinfo)
	{	errprintf("a parameter is NULL\n");
		return -1;
	}
	
	pthread_mutex_lock(&q->mutex);
	
	if (q->head==NULL)
	{	pthread_mutex_unlock(&q->mutex);
		errprintf("q->head is NULL: list is empty\n");
		return -1; // item not found
	}
	
	// test the first item
	for (cur=q->head; cur!=NULL; cur=cur->next)
	{
		if (cur->itemnum==itemnum) // block found
		{
			oldsize=get_block_size(&cur->blkinfo);
			//oldsize=(cur->blkinfo.blkcompressed==true)?(cur->blkinfo.blkarsize):(cur->blkinfo.blkrealsize);
			newsize=get_block_size(blkinfo);
			//newsize=(blkinfo->blkcompressed==true)?(blkinfo->blkarsize):(blkinfo->blkrealsize);
			q->memused-=oldsize; // dec the old size
			q->memused+=newsize; // inc the new size
			cur->status=newstatus;
			cur->blkinfo=*blkinfo;
			pthread_mutex_unlock(&q->mutex);
			pthread_cond_broadcast(&q->cond);
			return 0;
		}
	}
	
	errprintf("queue_replace_block(): cannot find item %ld\n", (long)itemnum);
	pthread_mutex_unlock(&q->mutex);
	
	return -1; // not found
}

// the compression thread requires the first block which has not yet been compressed
s64 queue_get_first_block_todo(struct s_queue *q, struct s_blockinfo *blkinfo)
{
	struct s_queueitem *cur;
	s64 itemfound=-1;
	
	if (!q || !blkinfo)
	{	errprintf("a parameter is NULL\n");
		return -1;
	}
	
	pthread_mutex_lock(&q->mutex);
	
	do
	{	
		for (cur=q->head; (cur!=NULL) && (itemfound==-1); cur=cur->next)
		{
			if ((cur->type==QITEM_TYPE_BLOCK) && (cur->status==QITEM_STATUS_TODO))
			{
				*blkinfo=cur->blkinfo;
				cur->status=QITEM_STATUS_PROGRESS;
				itemfound=cur->itemnum;
			}
		}
		
		if (itemfound==-1)
		{
			if (get_abort()==true || get_finished()==true)
			{	
				pthread_mutex_unlock(&q->mutex);
				return 0; // "0" means not-found
			}
			else
			{	
				struct timespec t=get_timeout();
				pthread_cond_timedwait(&q->cond, &q->mutex, &t);
			}
		}
	} while (itemfound==-1);
	
	pthread_mutex_unlock(&q->mutex);
	pthread_cond_broadcast(&q->cond);
	
	return itemfound; // ">0" means item found
}

// the writer thread requires the first block of the queue if it ready to go
s64 queue_dequeue_first(struct s_queue *q, int *type, char *magic, struct s_dico **dico, struct s_blockinfo *blkinfo)
{
	struct s_queueitem *cur=NULL;
	s64 itemfound=-1;
	
	if (!q || !magic || !dico || !blkinfo)
	{	errprintf("a parameter is NULL\n");
		return -1;
	}
	
	pthread_mutex_lock(&q->mutex);
	
	do
	{
		if (((cur=q->head)!=NULL) && (cur->status==QITEM_STATUS_DONE))
		{	
			if (cur->type==QITEM_TYPE_BLOCK) // item to dequeue is a block
			{
				q->memused-=get_block_size(&cur->blkinfo);
				//q->memused-=(cur->blkinfo.blkcompressed==true)?(cur->blkinfo.blkarsize):(cur->blkinfo.blkrealsize);
				*type=cur->type;
				itemfound=cur->itemnum;
				*blkinfo=cur->blkinfo;
				q->head=cur->next;
				free(cur);
			}
			else if (cur->type==QITEM_TYPE_DICO) // item to dequeue is a dico
			{
				*dico=cur->dico;
				memcpy(magic, cur->magic, FSA_SIZEOF_MAGIC); // header contents
				*type=cur->type;
				itemfound=cur->itemnum;
				q->head=cur->next;
				free(cur);
			}
			else
			{
				errprintf("invalid item type in queue\n");
				pthread_mutex_unlock(&q->mutex);
				return -1;
			}
		}
		if (itemfound==-1)
		{
			if (get_abort()==true || get_finished()==true)
			{	pthread_mutex_unlock(&q->mutex);
				return 0; // "0" means not-found
			}
			else
			{	
				struct timespec t=get_timeout();
				pthread_cond_timedwait(&q->cond, &q->mutex, &t);
			}
		}
	} while (itemfound==-1);
	
	pthread_mutex_unlock(&q->mutex);
	pthread_cond_broadcast(&q->cond);
	return itemfound; // ">0" means item found
}

// runs with the mutex already locked
int queue_count_internal(struct s_queue *q)
{
	struct s_queueitem *cur;
	int count=0;
	
	if (!q)
	{	errprintf("q is NULL\n");
		return -1;
	}
	
	cur=q->head;
	while (cur!=NULL)
	{
		count++;
		cur=cur->next;
	}
	
	return count;
}

// runs with the mutex unlocked (external users)
int queue_count(struct s_queue *q)
{
	int count=0;
	
	if (!q)
	{	errprintf("q is NULL\n");
		return -1;
	}
	
	pthread_mutex_lock(&q->mutex);
	count = queue_count_internal(q);
	pthread_mutex_unlock(&q->mutex);
	
	return count;
}

// how many items in the queue have a particular status
int queue_count_status(struct s_queue *q, int status)
{
	struct s_queueitem *cur;
	int count=0;
	
	if (!q)
	{	errprintf("q is NULL\n");
		return -1;
	}
	
	pthread_mutex_lock(&q->mutex);
	cur=q->head;
	while (cur!=NULL)
	{
		if (status==QITEM_STATUS_NULL || cur->status==status)
			count++;
		cur=cur->next;
	}
	pthread_mutex_unlock(&q->mutex);
	
	return count;
}

int queue_add_header(struct s_queue *q, struct s_dico *d, char *magic)
{
	struct s_queueitem *item;
	struct s_queueitem *cur;
	
	if (!q || !d || !magic)
	{	dbgprintf(2, "a parameter is NULL\n");
		return -1;
	}
	
	if (get_abort()==true)
		return -1;
		
	// create the new item in memory
	item=malloc(sizeof(struct s_queueitem));
	if (!item)
	{	dbgprintf(2, "out of memory 1\n");
		return -1;
	}
	
	memcpy(item->magic, magic, FSA_SIZEOF_MAGIC);
	item->type=QITEM_TYPE_DICO;
	item->dico=d;
	item->status=QITEM_STATUS_DONE;
	item->next=NULL;
	
	pthread_mutex_lock(&q->mutex);
	
	// if queue is full, let the other threads remove items first
	while (q->memused > q->memmax)
	{
		struct timespec t=get_timeout();
		pthread_cond_timedwait(&q->cond, &q->mutex, &t);
	}
	
	item->itemnum=q->curitemnum++;
	if (q->head==NULL) // if list empty
	{
		q->head=item;
	}
	else // list not empty
	{	for (cur=q->head; (cur!=NULL) && (cur->next!=NULL); cur=cur->next);
		cur->next=item;
	}
	
	pthread_mutex_unlock(&q->mutex);
	pthread_cond_broadcast(&q->cond);
	
	return 0;
}

int queue_dequeue_header(struct s_queue *q, struct s_dico **d, char *magicbuf)
{
	struct s_queueitem *cur;
	
	if (!q || !d || !magicbuf)
	{	dbgprintf(2, "one parameter is NULL: q=%p and d=%p and magicbuf=%p\n", q, d, magicbuf);
		return -1;
	}
	
	pthread_mutex_lock(&q->mutex);
	
	if (get_abort()==true)
	{	pthread_mutex_unlock(&q->mutex);
		return -1;
	}
	
	while ((queue_count_internal(q)<1) && (get_abort()==false))
	{
		struct timespec t=get_timeout();
		pthread_cond_timedwait(&q->cond, &q->mutex, &t);
	}
	
	if (get_abort()==true)
	{	pthread_mutex_unlock(&q->mutex);
		return -1;
	}
	
	if ((cur=q->head)==NULL)
	{	pthread_mutex_unlock(&q->mutex);
		errprintf("dequeue - cannot dequeue since q->head is NULL\n");
		return -1; // not found
	}
	
	// test the first item
	if (cur->type==QITEM_TYPE_DICO)
	{
		q->head=cur->next;
		*d=cur->dico;
		memcpy(magicbuf, cur->magic, FSA_SIZEOF_MAGIC);
		free(cur);
		pthread_mutex_unlock(&q->mutex);
		pthread_cond_broadcast(&q->cond);
		return 0;
	}
	else
	{
		errprintf("dequeue - wrong type of data in the queue: wanted a dico\n");
		pthread_mutex_unlock(&q->mutex);
		pthread_cond_broadcast(&q->cond);
		return -1;  // ok but not found
	}
}

// say whether or not the first item is a block and is ready
int queue_is_first_block_ready(struct s_queue *q)
{
	struct s_queueitem *cur;

	if (q==NULL)
		return false; // not found
	else	
		return ((cur=q->head)!=NULL) && (cur->type==QITEM_TYPE_BLOCK) && (cur->status==QITEM_STATUS_DONE);
}

// the extract function wants to read headers from the queue
int queue_dequeue_block(struct s_queue *q, struct s_blockinfo *blkinfo)
{	
	struct s_queueitem *cur;
	
	if (!q || !blkinfo)
	{	errprintf("a parameter is NULL\n");
		return -1;
	}
	
	pthread_mutex_lock(&q->mutex);
	
	if (get_abort()==true)
	{	pthread_mutex_unlock(&q->mutex);
		return -1;
	}
	
	while ((queue_is_first_block_ready(q)!=true) && (get_abort()==false))
	{
		struct timespec t=get_timeout();
		pthread_cond_timedwait(&q->cond, &q->mutex, &t);
	}
	
	if (get_abort()==true)
	{	pthread_mutex_unlock(&q->mutex);
		return -1;
	}
	
	// test the first item
	cur=q->head;
	if ((cur->type==QITEM_TYPE_BLOCK) && (cur->status==QITEM_STATUS_DONE))
	{
		q->memused-=get_block_size(&cur->blkinfo);
		//q->memused-=(cur->blkinfo.blkcompressed==true)?(cur->blkinfo.blkarsize):(cur->blkinfo.blkrealsize);
		*blkinfo=cur->blkinfo;
		q->head=cur->next;
		free(cur);
		pthread_mutex_unlock(&q->mutex);
		pthread_cond_broadcast(&q->cond);
		return 0;
	}
	else
	{
		errprintf("dequeue - wrong type of data in the queue: wanted a block, found dico with magic=[%4s]\n", cur->magic);
		pthread_mutex_unlock(&q->mutex);
		pthread_cond_broadcast(&q->cond);
		return -1;  // ok but not found
	}
}

// returns true if the first item in queue is a dico or a block which is ready
int queue_is_first_item_ready(struct s_queue *q)
{
	struct s_queueitem *cur;
	int ret=false;
	
	if (!q)
	{	errprintf("q is NULL\n");
		return false; // not found
	}
	
	if ((cur=q->head)==NULL)
		ret=false; // list empty
	else if (cur->type==QITEM_TYPE_DICO)
		ret=true; // a dico is always ready
	else if ((cur->type==QITEM_TYPE_BLOCK) && (cur->status==QITEM_STATUS_DONE))
		ret=true; // a block which has been prepared is ready
	else // other cases: block not yet prepared
		ret=false;
	
	return ret;
}

// say what the next item which is ready in the queue is but do not remove it
int queue_check_next_item(struct s_queue *q, int *type, char *magic)
{
	struct s_queueitem *cur;

	if (!q || !type || !magic)
	{	errprintf("a parameter is NULL\n");
		return -1;
	}	
	
	memset(magic, 0, FSA_SIZEOF_MAGIC);
	*type=0;	
	
	pthread_mutex_lock(&q->mutex);
	
	if (get_abort()==true)
	{	pthread_mutex_unlock(&q->mutex);
		return -1;
	}
	
	while ((queue_is_first_item_ready(q)==false) && (get_abort()==false))
	{
		struct timespec t=get_timeout();
		pthread_cond_timedwait(&q->cond, &q->mutex, &t);
	}
	
	if (get_abort()==true)
	{	pthread_mutex_unlock(&q->mutex);
		return -1;
	}
	
	// test the first item
	if (((cur=q->head)!=NULL) && (cur->status==QITEM_STATUS_DONE))
	{
		if (cur->type==QITEM_TYPE_BLOCK) // item to dequeue is a block
		{
			*type=cur->type;
			memset(magic, 0, FSA_SIZEOF_MAGIC);
			pthread_mutex_unlock(&q->mutex);
			return 0;
		}
		else if (cur->type==QITEM_TYPE_DICO) // item to dequeue is a dico
		{
			memcpy(magic, cur->magic, FSA_SIZEOF_MAGIC); // header contents
			*type=cur->type;
			pthread_mutex_unlock(&q->mutex);
			return 0;
		}
		else
		{
			errprintf("invalid item type in queue\n");
			return -1;
		}
	}
	
	pthread_mutex_unlock(&q->mutex);
	pthread_cond_broadcast(&q->cond);
	return -1;  // not found
}

// destroy the first item in the queue (similar to dequeue but do not read it)
int queue_destroy_first_item(struct s_queue *q)
{
	struct s_queueitem *cur;
	
	if (!q)
	{	errprintf("q is NULL\n");
		return -1;
	}
	
	pthread_mutex_lock(&q->mutex);
	if ((cur=q->head)==NULL)
	{	pthread_mutex_unlock(&q->mutex);
		errprintf("dequeue - cannot dequeue since q->head is NULL\n");
		return -1; // not found
	}
	
	if (cur->status==QITEM_TYPE_BLOCK)
		q->memused-=get_block_size(&cur->blkinfo);
		//q->memused-=(cur->blkinfo.blkcompressed==true)?(cur->blkinfo.blkarsize):(cur->blkinfo.blkrealsize);
	q->head=cur->next;
	free(cur);
	pthread_mutex_unlock(&q->mutex);
	pthread_cond_broadcast(&q->cond);
	return 0;
}

int queue_print(struct s_queue *q)
{
	struct s_queueitem *cur;
	int count=0;

	if (!q)
	{	errprintf("q is NULL\n");
		return -1;
	}

	pthread_mutex_lock(&q->mutex);
	printf("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - \n");
	cur=q->head;
	while (cur!=NULL)
	{
		printf("item[%.2d] has type=%d\n", count++, cur->type);
		cur=cur->next;
	}
	printf("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - \n");
	pthread_mutex_unlock(&q->mutex);
	
	return count;
}
