/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#ifndef __QUEUE_H__
#define __QUEUE_H__

#include <pthread.h>

enum	{QITEM_STATUS_NULL=0, QITEM_STATUS_TODO, QITEM_STATUS_PROGRESS, QITEM_STATUS_DONE};
enum	{QITEM_TYPE_NULL=0, QITEM_TYPE_BLOCK, QITEM_TYPE_DICO};

struct s_dico;

struct s_blockinfo
{
	char	*blkdata; // pointer to data block (compressed or uncompressed)
	u32	blkrealsize; // size of the data in the normal state (uncompressed)
	u64	blkoffset; // offset of the block in the normal file
	u32	blkarcsum; // checksum of the block as it it when it's in the archive (compressed)
	u64	blkarsize; // size of the block as it is in the archive (compressed)
	u32	blkcompalgo; // algo used to compressed the block
	bool	blklocked; // true if locked (being compressed)
	bool	blkcompressed; // true if the block has been processed by the compression thread
};

struct s_queueitem
{	// ------- common information: for all items of the queue --------------------------
	int 			type; // QITEM_TYPE_BLOCK or QITEM_TYPE_DICO
	s64			itemnum; // unique identifier of the item in the queue
	int			status; // compressed, being-compressed, not-yet-compressed
	struct s_queueitem	*next; // next block in the linked list
	// ------- information used when type==QITEM_TYPE_BLOCK (for blocks only) ----------
	struct s_blockinfo 	blkinfo; // extra info used for compressed blocks
	// ------- information used when type==QITEM_TYPE_DICO (for headers only) ----------
	char			magic[FSA_SIZEOF_MAGIC+1]; // magic for dico
	struct s_dico		*dico; // if (type==QITEM_TYPE_DICO)
};

struct s_queue
{
	struct s_queueitem	*head; // head of the queue: first item
	pthread_mutex_t		mutex; // pthread mutex for data protection
	pthread_cond_t		cond; // condition for pthread synchronization
	s64			memused; // number of bytes that have already been allocated for data-blocks
	s64			memmax; // number of bytes that the queue can allocate for data-blocks
	s64			curitemnum; // unique id given to every new item (block or header)
};

int queue_init(struct s_queue *l, s64 memmax);
int queue_destroy(struct s_queue *l);
int queue_count(struct s_queue *l);
int queue_count_internal(struct s_queue *q);
int queue_count_status(struct s_queue *l, int status);
s64 queue_get_first_block_todo(struct s_queue *q, struct s_blockinfo *blkinfo);
int queue_replace_block(struct s_queue *q, s64 itemnum, struct s_blockinfo *blkinfo, int newstatus);
int queue_add_block(struct s_queue *q, struct s_blockinfo *blkinfo, int status);
int queue_add_header(struct s_queue *q, struct s_dico *d, char *magic);
int queue_dequeue_header(struct s_queue *q, struct s_dico **d, char *magicbuf);
int queue_dequeue_block(struct s_queue *q, struct s_blockinfo *blkinfo);
s64 queue_dequeue_first(struct s_queue *q, int *type, char *magic, struct s_dico **dico, struct s_blockinfo *blkinfo);
int queue_is_first_item_ready(struct s_queue *q);
int queue_check_next_item(struct s_queue *q, int *type, char *magic);
int queue_destroy_first_item(struct s_queue *q);
int queue_print(struct s_queue *l);

#endif // __QUEUE_H__
